#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Module for Monte Carlo generation of sequences from a V(D)J recomb model.

    Copyright (C) 2018 Zachary Sethna

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.


This module defines classes to randomly sample a V(D)J recomb model and
assemble CDR3 sequences defined by the sampled recombination events.



The method gen_rnd_prod_CDR3 will generate the actual CDR3 sequences and
outputs both the nucleotide sequence and the amino acid sequence. It will also
output the V and J gene/allele choices which can be discarded or stored as
the user desires. If a 'sequence read' (i.e. a sequence which mimics the
primer positions from an actual sequencing experiment) is needed, the V and J
identities can be used to determine what the sequence is outside of the CDR3
region.

Example
-------
>>> import olga.load_model as load_model
>>> import olga.sequence_generation as seq_gen
>>>
>>> params_file_name = './models/human_T_beta/model_params.txt'
>>> marginals_file_name = './models/human_T_beta/model_marginals.txt'
>>> V_anchor_pos_file ='./models/human_T_beta/V_gene_CDR3_anchors.csv'
>>> J_anchor_pos_file = './models/human_T_beta/J_gene_CDR3_anchors.csv'
>>>
>>> genomic_data = load_model.GenomicDataVDJ()
>>> genomic_data.load_igor_genomic_data(params_file_name, V_anchor_pos_file, J_anchor_pos_file)
>>>
>>> generative_model = load_model.GenerativeModelVDJ()
>>> generative_model.load_and_process_igor_model(marginals_file_name)
>>>
>>> seq_gen_model = seq_gen.SequenceGenerationVDJ(generative_model, genomic_data)
>>>
>>> seq_gen_model.gen_rnd_prod_CDR3()
('TGTGCCAGCAGTGAAAAAAGGCAATGGGAAAGCGGGGAGCTGTTTTTT', 'CASSEKRQWESGELFF', 27, 8)
>>> seq_gen_model.gen_rnd_prod_CDR3()
('TGTGCCAGCAGTTTAGTGGGAAGGGCGGGGCCCTATGGCTACACCTTC', 'CASSLVGRAGPYGYTF', 14, 1)
>>> seq_gen_model.gen_rnd_prod_CDR3()
('TGTGCCAGCTGGACAGGGGGCAACTACGAGCAGTACTTC', 'CASWTGGNYEQYF', 55, 13)

@author: zacharysethna
"""
from __future__ import division
import numpy as np
from olga.utils import nt2aa, calc_steady_state_dist

class SequenceGenerationVDJ(object):
    """Class of to generate sequences from a VDJ generative model.

    Attributes
    ----------
    CPV : ndarray
        Cumulative probability distribution of V usage
    given_V_CPdelV : ndarray
        Conditional cumulative distribution of the V deletions given V.

    CPDJ : ndarray
        Joint cumulative probability distribution of the D and J usages.
    given_J_CPdelJ : ndarray
        Conditional cumulative distribution of the J deletions given the J.

    given_D_CPdelDldelDr : ndarray
        Joint cumulative probability distribution of the D deletions given the
        D.

    CPinsVD : ndarray
        Cumulative probability distribution of the VD (N1) insertion sequence
        length.
    CPinsDJ : ndarray
        Cumulative probability distribution of the DJ (N2) insertion sequence
        length.
    CRvd : ndarray
        Cumulative Markov transition matrix for the VD insertion junction.
    CRdj : ndarray
        Cumulative Markov transition matrix for the DJ insertion junction.
    C_first_nt_bias_insVD : ndarray
        (4,) array of the cumulative probability distribution of the indentity
        of the first nucleotide insertion for the VD junction.
    C_first_nt_bias_insDJ : ndarray
        (4,) array of the cumulative probability distribution of the indentity
        of the first nucleotide insertion for the DJ junction.

    num_J_genes : int
        Number of J genes/alleles.
    num_delDr_poss : int
        Number of delDr possibilities.

    cutV_genomic_CDR3_segs : list of strings
        List of the V germline nucleotide sequences, trimmed to begin at the
        CDR3 region (includes the conserved C residue) with the maximum number
        of reverse complementary palindromic insertions appended.
    cutD_genomic_CDR3_segs : list of strings
        List of the D germline nucleotide sequences, with the maximum number
        of reverse complementary palindromic insertions appended to both ends.
    cutJ_genomic_CDR3_segs : list of strings
        List of the J germline nucleotide sequences, trimmed to end at the
        CDR3 region (includes the conserved F or W residue) with the maximum
        number of reverse complementary palindromic insertions appended.

    """

    def __init__(self, generative_model, genomic_data):
        """Initialize SequenceGenerationVDJ

        This intialization computes all of the cumulative probability
        distributions that will be needed for efficient Monte Carlo sequence
        generation out of a GenerativeModelVDJ.

        Parameters
        ----------
        generative_model : GenerativeModelVDJ
            VDJ generative model class containing the model parameters.
        genomic_data : GenomicDataVDJ
            VDJ genomic data class containing the V, D, and J germline
            sequences and info.

        """

        self.CPV = (generative_model.PV/np.sum(generative_model.PV)).cumsum()
        self.CPDJ = (generative_model.PDJ/np.sum(generative_model.PDJ)).flatten().cumsum()
        self.CinsVD = (generative_model.PinsVD/np.sum(generative_model.PinsVD)).cumsum()
        self.CinsDJ = (generative_model.PinsDJ/np.sum(generative_model.PinsDJ)).cumsum()


        for V in range(generative_model.PdelV_given_V.shape[1]):
            if np.sum(generative_model.PdelV_given_V[:, V])> 0:
                generative_model.PdelV_given_V[:, V] = generative_model.PdelV_given_V[:, V]/np.sum(generative_model.PdelV_given_V[:, V])

        self.given_V_CPdelV = generative_model.PdelV_given_V.T.cumsum(axis = 1)

        for J in range(generative_model.PdelJ_given_J.shape[1]):
            if np.sum(generative_model.PdelJ_given_J[:, J])> 0:
                generative_model.PdelJ_given_J[:, J] = generative_model.PdelJ_given_J[:, J]/np.sum(generative_model.PdelJ_given_J[:, J])

        self.given_J_CPdelJ = generative_model.PdelJ_given_J.T.cumsum(axis = 1)

        for D in range(generative_model.PdelDldelDr_given_D.shape[2]):
            if np.sum(generative_model.PdelDldelDr_given_D[:,  :, D]) > 0:
                generative_model.PdelDldelDr_given_D[:, :, D] = generative_model.PdelDldelDr_given_D[:, :, D]/np.sum(generative_model.PdelDldelDr_given_D[:, :, D])

        self.given_D_CPdelDldelDr = np.array([ generative_model.PdelDldelDr_given_D[:, :, i].flatten().cumsum() for i in range(generative_model.PdelDldelDr_given_D.shape[2])])


        self.C_Rvd = generative_model.Rvd.T.cumsum(axis = 1)
        self.C_Rdj = generative_model.Rdj.T.cumsum(axis = 1)

        if generative_model.first_nt_bias_insVD is None:
            first_nt_bias_insVD = calc_steady_state_dist(generative_model.Rvd)
        else:
            first_nt_bias_insVD = generative_model.first_nt_bias_insVD

        if generative_model.first_nt_bias_insDJ is None:
            first_nt_bias_insDJ = calc_steady_state_dist(generative_model.Rdj)
        else:
            first_nt_bias_insDJ = generative_model.first_nt_bias_insDJ



        self.C_first_nt_bias_insVD = first_nt_bias_insVD.cumsum()
        self.C_first_nt_bias_insDJ = first_nt_bias_insDJ.cumsum()

        self.num_J_genes = generative_model.PDJ.shape[1]
        self.num_delDr_poss = generative_model.PdelDldelDr_given_D.shape[1]


        self.cutV_genomic_CDR3_segs = genomic_data.cutV_genomic_CDR3_segs
        self.cutD_genomic_CDR3_segs = genomic_data.cutD_genomic_CDR3_segs
        self.cutJ_genomic_CDR3_segs = genomic_data.cutJ_genomic_CDR3_segs

    def gen_rnd_prod_CDR3(self, conserved_J_residues = 'FVW'):
        """Generate a productive CDR3 seq from a Monte Carlo draw of the model.

        Parameters
        ----------
        conserved_J_residues : str, optional
            Conserved amino acid residues defining the CDR3 on the J side (normally
            F, V, and/or W)

        Returns
        -------
        ntseq : str
            Productive CDR3 nucleotide sequence
        aaseq : str
            CDR3 amino acid sequence (aaseq = nt2aa(ntseq))
        V_choice : int
            Index of V allele chosen to generate the CDR3 seq
        J_choice : int
            Index of J allele chosen to generate the CDR3 seq

        """
        coding_pass = False

        while not coding_pass:
            recomb_events = self.choose_random_recomb_events()
            V_seq = self.cutV_genomic_CDR3_segs[recomb_events['V']]

            #This both checks that the position of the conserved C is
            #identified and that the V isn't fully deleted out of the CDR3
            #region
            if len(V_seq) <= max(recomb_events['delV'], 0):
                continue
            D_seq = self.cutD_genomic_CDR3_segs[recomb_events['D']]
            J_seq = self.cutJ_genomic_CDR3_segs[recomb_events['J']]

            #We check that the D and J aren't deleted more than allowed. Note
            #the generative model really should reflect this structure already
            if len(D_seq) < (recomb_events['delDl'] + recomb_events['delDr']) or len(J_seq) < recomb_events['delJ']:
                continue

            V_seq = V_seq[:len(V_seq) - recomb_events['delV']]
            D_seq = D_seq[recomb_events['delDl']:len(D_seq)-recomb_events['delDr']]
            J_seq = J_seq[recomb_events['delJ']:]

            if (len(V_seq)+ len(D_seq) + len(J_seq) + recomb_events['insVD'] + recomb_events['insDJ']) % 3 != 0:
                continue


            insVD_seq = rnd_ins_seq(recomb_events['insVD'], self.C_Rvd, self.C_first_nt_bias_insVD)
            insDJ_seq = rnd_ins_seq(recomb_events['insDJ'], self.C_Rdj, self.C_first_nt_bias_insDJ)[::-1] #have to reverse the DJ seq

            #Translate to amino acid sequence, see if productive
            ntseq = V_seq + insVD_seq + D_seq + insDJ_seq + J_seq
            aaseq = nt2aa(ntseq)

            if '*' not in aaseq and aaseq[0]=='C' and aaseq[-1] in conserved_J_residues:
                return ntseq, aaseq, recomb_events['V'], recomb_events['J']

    def choose_random_recomb_events(self):
        """Sample the genomic model for VDJ recombination events.

        Returns
        -------
        recomb_events : dict
            Dictionary of the VDJ recombination events. These are
            integers determining gene choice, deletions, and number of insertions.

        Example
        --------
        >>> sequence_generation.choose_random_recomb_events()
        {'D': 0, 'J': 13, 'V': 36, 'delDl': 2, 'delDr': 13, 'delJ': 10, 'delV': 5, 'insDJ': 6, 'insVD': 9}

        """

        recomb_events = {}
        recomb_events['V'] = self.CPV.searchsorted(np.random.random())

        #For 2D arrays make sure to take advantage of a mod expansion to find indicies
        DJ_choice = self.CPDJ.searchsorted(np.random.random())
        recomb_events['D'] = DJ_choice//self.num_J_genes
        recomb_events['J'] = DJ_choice % self.num_J_genes


        #Refer to the correct slices for the dependent distributions
        recomb_events['delV'] = self.given_V_CPdelV[recomb_events['V'], :].searchsorted(np.random.random())

        recomb_events['delJ'] = self.given_J_CPdelJ[recomb_events['J'], :].searchsorted(np.random.random())

        delDldelDr_choice = self.given_D_CPdelDldelDr[recomb_events['D'], :].searchsorted(np.random.random())

        recomb_events['delDl'] = delDldelDr_choice//self.num_delDr_poss
        recomb_events['delDr'] = delDldelDr_choice % self.num_delDr_poss

        recomb_events['insVD'] = self.CinsVD.searchsorted(np.random.random())
        recomb_events['insDJ'] = self.CinsDJ.searchsorted(np.random.random())

        return recomb_events

#%%
class SequenceGenerationVJ(object):
    """Class of to generate sequences from a VJ generative model.

    Attributes
    ----------
    CPVJ : ndarray
        Joint cumulative probability distribution of the V and J usages.
    CPdelV_given_V : ndarray
        Conditional cumulative distribution of the V deletions given the V.
    CPdelJ_given_J : ndarray
        Conditional cumulative distribution of the J deletions given the J.

    CPinsVJ : ndarray
        Cumulative probability distribution of the VJ (N) insertion sequence
        length.
    CRvj : ndarray
        Cumulative Markov transition matrix for the VJ insertion junction.
    C_first_nt_bias_insVJ : ndarray
        (4,) array of the cumulative probability distribution of the indentity
        of the first nucleotide insertion for the VD junction.

    num_J_genes : int
        Number of J genes/alleles

    cutV_genomic_CDR3_segs : list of strings
        List of the V germline nucleotide sequences, trimmed to begin at the
        CDR3 region (includes the conserved C residue) with the maximum number
        of reverse complementary palindromic insertions appended.
    cutJ_genomic_CDR3_segs : list of strings
        List of the J germline nucleotide sequences, trimmed to end at the
        CDR3 region (includes the conserved F or W residue) with the maximum
        number of reverse complementary palindromic insertions appended.

    """

    def __init__(self, generative_model, genomic_data):
        """Initialize SequenceGenerationVJ

        This intialization computes all of the cumulative probability
        distributions that will be needed for efficient Monte Carlo sequence
        generation out of a GenerativeModelVJ.

        Parameters
        ----------
        generative_model : GenerativeModelVJ
            VJ generative model class containing the model parameters.
        genomic_data : GenomicDataVJ
            VJ genomic data class containing the V and J germline
            sequences and info.

        """

        self.CPVJ = (generative_model.PVJ/np.sum(generative_model.PVJ)).flatten().cumsum()
        self.CPinsVJ = (generative_model.PinsVJ/np.sum(generative_model.PinsVJ)).cumsum()

        for V in range(generative_model.PdelV_given_V.shape[1]):
            if np.sum(generative_model.PdelV_given_V[:, V])> 0:
                generative_model.PdelV_given_V[:, V] = generative_model.PdelV_given_V[:, V]/np.sum(generative_model.PdelV_given_V[:, V])

        self.given_V_CPdelV = generative_model.PdelV_given_V.T.cumsum(axis = 1)

        for J in range(generative_model.PdelJ_given_J.shape[1]):
            if np.sum(generative_model.PdelJ_given_J[:, J])> 0:
                generative_model.PdelJ_given_J[:, J] = generative_model.PdelJ_given_J[:, J]/np.sum(generative_model.PdelJ_given_J[:, J])

        self.given_J_CPdelJ = generative_model.PdelJ_given_J.T.cumsum(axis = 1)


        self.C_Rvj = generative_model.Rvj.T.cumsum(axis = 1)

        if generative_model.first_nt_bias_insVJ == None:
            first_nt_bias_insVJ = calc_steady_state_dist(generative_model.Rvj)
        else:
            first_nt_bias_insVJ = generative_model.first_nt_bias_insVJ

        self.C_first_nt_bias_insVJ = first_nt_bias_insVJ.cumsum()

        self.num_J_genes = generative_model.PVJ.shape[1]

        self.cutV_genomic_CDR3_segs = genomic_data.cutV_genomic_CDR3_segs
        self.cutJ_genomic_CDR3_segs = genomic_data.cutJ_genomic_CDR3_segs

    def gen_rnd_prod_CDR3(self, conserved_J_residues = 'FVW'):
        """Generate a productive CDR3 seq from a Monte Carlo draw of the model.

        Parameters
        ----------
        conserved_J_residues : str, optional
            Conserved amino acid residues defining the CDR3 on the J side (normally
            F, V, and/or W)

        Returns
        -------
        ntseq : str
            Productive CDR3 nucleotide sequence
        aaseq : str
            CDR3 amino acid sequence (aaseq = nt2aa(ntseq))
        V_choice : int
            Index of V allele chosen to generate the CDR3 seq
        J_choice : int
            Index of J allele chosen to generate the CDR3 seq

        """

        coding_pass = False

        while not coding_pass:
            recomb_events = self.choose_random_recomb_events()
            V_seq = self.cutV_genomic_CDR3_segs[recomb_events['V']]

            #This both checks that the position of the conserved C is
            #identified and that the V isn't fully deleted out of the CDR3
            #region
            if len(V_seq) <= max(recomb_events['delV'], 0):
                continue
            J_seq = self.cutJ_genomic_CDR3_segs[recomb_events['J']]

            #We check that J isn't deleted more than allowed. Note the
            #generative model really should reflect this structure already
            if len(J_seq) < recomb_events['delJ']:
                continue

            V_seq = V_seq[:len(V_seq) - recomb_events['delV']]
            J_seq = J_seq[recomb_events['delJ']:]

            if (len(V_seq)+len(J_seq) + recomb_events['insVJ']) % 3 != 0:
                continue


            insVJ_seq = rnd_ins_seq(recomb_events['insVJ'], self.C_Rvj, self.C_first_nt_bias_insVJ)

            #Translate to amino acid sequence, see if productive
            ntseq = V_seq + insVJ_seq + J_seq
            aaseq = nt2aa(ntseq)

            if '*' not in aaseq and aaseq[0]=='C' and aaseq[-1] in conserved_J_residues:
                return ntseq, aaseq, recomb_events['V'], recomb_events['J']

    def choose_random_recomb_events(self):
        """Sample the genomic model for VDJ recombination events.

        Returns
        -------
        recomb_events : dict
            Dictionary of the VDJ recombination events. These are
            integers determining gene choice, deletions, and number of insertions.

        Example
        --------
        >>> sequence_generation.choose_random_recomb_events()
        {'J': 13, 'V': 36, 'delJ': 10, 'delV': 5, 'insVJ': 3}

        """

        recomb_events = {}

        #For 2D arrays make sure to take advantage of a mod expansion to find indicies
        VJ_choice = self.CPVJ.searchsorted(np.random.random())
        recomb_events['V'] = VJ_choice//self.num_J_genes
        recomb_events['J'] = VJ_choice % self.num_J_genes


        #Refer to the correct slices for the dependent distributions
        recomb_events['delV'] = self.given_V_CPdelV[recomb_events['V'], :].searchsorted(np.random.random())

        recomb_events['delJ'] = self.given_J_CPdelJ[recomb_events['J'], :].searchsorted(np.random.random())
        recomb_events['insVJ'] = self.CPinsVJ.searchsorted(np.random.random())

        return recomb_events

#%% Function to get sequence identity (of a given length)
def rnd_ins_seq(ins_len, C_R, CP_first_nt):
    """Generate a random insertion nucleotide sequence of length ins_len.

    Draws the sequence identity (for a set length) from the distribution
    defined by the dinucleotide markov model of transition matrix R.

    Parameters
    ----------
    ins_len : int
        Length of nucleotide sequence to be inserted.
    C_R : ndarray
        (4, 4) array of the cumulative transition probabilities defined by the
        Markov transition matrix R
    CP_first_nt : ndarray
        (4,) array of the cumulative probabilities for the first inserted
        nucleotide

    Returns
    -------
    seq : str
        Randomly generated insertion sequence of length ins_len.

    Examples
    --------
    >>> rnd_ins_seq(7, CP_generative_model['C_Rvd'], CP_generative_model['C_first_nt_bias_insVD'])
    'GATGGAC'
    >>> rnd_ins_seq(7, CP_generative_model['C_Rvd'], CP_generative_model['C_first_nt_bias_insVD'])
    'ACCCCCG'
    >>> rnd_ins_seq(3, CP_generative_model['C_Rvd'], CP_generative_model['C_first_nt_bias_insVD'])
    'GCC'

    """
    nt2num = {'A': 0, 'C': 1, 'G': 2, 'T': 3}
    num2nt = 'ACGT'

    if ins_len == 0:
        return ''

    seq = num2nt[CP_first_nt.searchsorted(np.random.random())]
    ins_len += -1

    while ins_len > 0:
        seq += num2nt[C_R[nt2num[seq[-1]], :].searchsorted(np.random.random())]
        ins_len += -1

    return seq



import numba as nb
import numpy as np


@nb.njit(cache=True, fastmath=True)
def compute_Pi_L_numba(
    out: np.ndarray,
    aa_idx_seq: np.ndarray,
    Pixy: np.ndarray,
    Pins: np.ndarray,
    fnb: np.ndarray,
    znb: np.ndarray,
    S: np.ndarray,
    D: np.ndarray,
    T: np.ndarray,
    lD: np.ndarray,
    lT: np.ndarray,
    max_align: int
) -> None:
    
    max_insertions = Pins.size - 1
    M_codons = max_insertions // 3

    v0 = np.empty(4)
    cur = np.empty(4)

    _fill_from_left(
        out, aa_idx_seq, Pixy, Pins, fnb, znb,
        S, D, T, lD, lT, max_align, v0, cur,
        M_codons
    )


@nb.njit(cache=True, fastmath=True)
def compute_Pi_V_insVJ_given_J_numba(
    out: np.ndarray,
    aa_idx_seq: np.ndarray,
    Pixy: np.ndarray,
    Pins: np.ndarray,
    fnb: np.ndarray,
    znb: np.ndarray,
    S: np.ndarray,
    D: np.ndarray,
    T: np.ndarray,
    lD: np.ndarray,
    lT: np.ndarray,
    max_align: int
) -> None:

    numD = Pixy.shape[0]
    max_insertions = Pins.size - 1
    M_codons = max_insertions // 3

    v0 = np.empty(4)
    cur = np.empty(4)

    for d in range(numD):
        O = out[d]
        Pij = Pixy[d]

        _fill_from_left(
            O, aa_idx_seq, Pij, Pins, fnb, znb, 
            S, D, T, lD, lT, max_align, v0, cur, 
            M_codons
        )


@nb.njit(fastmath=True)
def _fill_from_left(
    O: np.ndarray,
    aa_idx_seq: np.ndarray,
    Pij: np.ndarray,
    Pins: np.ndarray,
    fnb: np.ndarray,
    znb: np.ndarray,
    S: np.ndarray,
    D: np.ndarray,
    T: np.ndarray,
    lD: np.ndarray,
    lT: np.ndarray,
    max_align: int,
    v0: np.ndarray,
    cur: np.ndarray,
    M_codons: int
) -> None:

    # Frame 1
    j0 = 0
    for init_pos in range(0, max_align, 3):
        _load4(Pij, v0, init_pos)
        _axpy4(Pins[0], v0, O, init_pos)

        aa0 = aa_idx_seq[j0]
        _dot4_axpy4(Pins[1], lD[aa0], v0, O, init_pos + 1)

        _dot4(lT[aa0], v0, cur)
        s = _sum4(cur)
        O[0, init_pos + 2] += Pins[2] * s

        base_ins = 2
        _walk_from_left(
            base_ins, j0, M_codons, aa_idx_seq,
            O, init_pos, Pins, S, D, T, cur
        )
        j0 += 1

    # Frame 2
    j0 = 0
    for init_pos in range(1, max_align, 3):
        _load4(Pij, v0, init_pos)
        _axpy4(Pins[0], v0, O, init_pos)
    
        for k in range(4):
            cur[k] = v0[k] * fnb[k]

        s = _sum4(cur)
        O[0, init_pos + 1] += Pins[1] * s

        base_ins = 1
        _walk_from_left(
            base_ins, j0, M_codons, aa_idx_seq,
            O, init_pos, Pins, S, D, T, cur
        )
        j0 += 1

    # Frame 3
    j0 = 0
    for init_pos in range(2, max_align, 3):
        a0 = Pij[0, init_pos]
        O[0, init_pos] += Pins[0] * a0

        for k in range(4):
            cur[k] = znb[k] * a0

        base_ins = 0
        _walk_from_left(
            base_ins, j0, M_codons, aa_idx_seq,
            O, init_pos, Pins, S, D, T, cur
        )
        j0 += 1


@nb.njit(cache=True, fastmath=True)
def compute_Pi_JinsDJ_given_D_numba(
    out: np.ndarray,
    aa_idx_seq: np.ndarray,
    Pixy: np.ndarray,
    Pins: np.ndarray,
    fnb: np.ndarray,
    znb: np.ndarray,
    S: np.ndarray,
    D: np.ndarray,
    T: np.ndarray,
    rD: np.ndarray,
    rT: np.ndarray,
    max_align: int
) -> None:

    L = aa_idx_seq.shape[0]
    numD = Pixy.shape[0]
    max_insertions = Pins.size - 1
    M_codons = max_insertions // 3

    v0 = np.empty(4)
    cur = np.empty(4)

    for d in range(numD):
        O = out[d]
        Pij = Pixy[d]

        # Frame 1
        j0 = L - 1
        for init_pos in range(-1, -(max_align + 1), -3):
            _load4(Pij, v0, init_pos)
            _axpy4(Pins[0], v0, O, init_pos)

            aa0 = aa_idx_seq[j0]
            _dot4_axpy4(Pins[1], rD[aa0], v0, O, init_pos - 1)
            _dot4(rT[aa0], v0, cur)

            s = _sum4(cur)
            O[0, init_pos - 2] += Pins[2] * s
            base_ins = 2
            _walk_from_right(base_ins, j0, M_codons, aa_idx_seq, O, init_pos, Pins, S, D, T, cur)
            j0 -= 1

        # Frame 2
        j0 = L - 1
        for init_pos in range(-2, -(max_align + 1), -3):
            _load4(Pij, v0, init_pos)
            _axpy4(Pins[0], v0, O, init_pos)

            for k in range(4):
                cur[k] = v0[k] * fnb[k]

            s = _sum4(cur)
            O[0, init_pos - 1] += Pins[1] * s
            base_ins = 1
            _walk_from_right(base_ins, j0, M_codons, aa_idx_seq, O, init_pos, Pins, S, D, T, cur)
            j0 -= 1

        # Frame 3
        j0 = L - 1
        for init_pos in range(-3, -(max_align + 1), -3):
            a0 = Pij[0, init_pos]
            O[0, init_pos] += Pins[0] * a0

            for k in range(4):
                cur[k] = znb[k] * a0

            base_ins = 0
            _walk_from_right(base_ins, j0, M_codons, aa_idx_seq, O, init_pos, Pins, S, D, T, cur)
            j0 -= 1


@nb.njit(fastmath=True)
def _walk_from_right(
    base_ins: int,
    j0: int,
    M_codons: int,
    aa_idx_seq: np.ndarray,
    O: np.ndarray,
    init_pos: int,
    Pins: np.ndarray,
    S: np.ndarray,
    D: np.ndarray,
    T: np.ndarray,
    cur: np.ndarray
) -> None:
    start_j = j0 - 1
    end_j_inc = max(0, j0 - M_codons + 1)
    for j in range(start_j, end_j_inc - 1, -1):
        aidx = aa_idx_seq[j]

        p0 = base_ins + 1
        p = init_pos - base_ins - 1
        _dot4_axpy4(Pins[p0], S[aidx], cur, O, p)

        p0 += 1
        p = init_pos - base_ins - 2
        _dot4_axpy4(Pins[p0], D[aidx], cur, O, p)

        _dot4_inplace(T[aidx], cur)

        p = init_pos - base_ins - 3
        s = _sum4(cur)
        O[0, p] += Pins[base_ins + 3] * s
        base_ins += 3


@nb.njit(fastmath=True)
def _walk_from_left(
    base_ins: int,
    j0: int,
    M_codons: int,
    aa_idx_seq: np.ndarray,
    O: np.ndarray,
    init_pos: int,
    Pins: np.ndarray,
    S: np.ndarray,
    D: np.ndarray,
    T: np.ndarray,
    cur: np.ndarray
) -> None:
    L = aa_idx_seq.shape[0]
    start_j = j0 + 1
    end_j = min(L, j0 + M_codons)

    for j in range(start_j, end_j):
        aidx = aa_idx_seq[j]

        p0 = base_ins + 1
        p = init_pos + p0
        _dot4_axpy4(Pins[p0], S[aidx], cur, O, p)

        p0 += 1
        p = init_pos + p0
        _dot4_axpy4(Pins[p0], D[aidx], cur, O, p)

        _dot4_inplace(T[aidx], cur)
        p = init_pos + base_ins + 3
        s = _sum4(cur)
        O[0, p] += Pins[base_ins + 3] * s
        base_ins += 3


@nb.njit(cache=True, fastmath=True)
def compute_Pi_R_one_numba(
    Pi_R: np.ndarray,
    L3: int,
    min_pos: int,
    dseq: np.ndarray,
    Pij: np.ndarray,
    Pdel_D: np.ndarray,
    max_del_D: np.ndarray,
    min_del_D: np.ndarray,
    PD_nt: np.ndarray,
    PD_2nd_stack: np.ndarray,
    zeroD: float,
    aa_idx_seq: np.ndarray,
    allow_lsf_stack,
    allow1: np.ndarray,
    allow2: np.ndarray,
    allow3: np.ndarray
) -> None:
    Ld = dseq.shape[0]
    num_delr = Pdel_D.shape[1]

    base_pair_prob = np.empty((4, 4))
    v0 = np.empty(4)

    # Frame 1
    for init_pos in range(-1, -L3 - 1, -3):
        _load4(Pij, v0, init_pos)
        for k in range(4):
            Pi_R[k, init_pos] += zeroD * v0[k]

        aa0 = aa_idx_seq[(init_pos + L3) // 3]
        lsf = allow_lsf_stack[aa0]

        for l in range(4):
            for s in range(4):
                acc = 0.0
                for t in range(4):
                    if lsf[l, s, t]:
                        acc += v0[t]
                base_pair_prob[l, s] = acc

        for delDr in range(num_delr):
            if min_del_D[delDr] == -1:
                continue
            
            dpos1 = Ld - delDr - 1
            if dpos1 <= max_del_D[delDr]:
                second_idx = dseq[dpos1]
                w = Pdel_D[dpos1, delDr]
                Pi_R[0, init_pos - 1] += w * base_pair_prob[0, second_idx]
                Pi_R[1, init_pos - 1] += w * base_pair_prob[1, second_idx]
                Pi_R[2, init_pos - 1] += w * base_pair_prob[2, second_idx]
                Pi_R[3, init_pos - 1] += w * base_pair_prob[3, second_idx]

            dpos2 = Ld - delDr - 2
            
            last_idx = dseq[dpos2]
            second_idx = dseq[dpos2 + 1]
            base_prob = base_pair_prob[last_idx, second_idx]
            if base_prob == 0.0:
                continue

            if dpos2 <= max_del_D[delDr]:
                Pi_R[0, init_pos - 2] += Pdel_D[dpos2, delDr] * base_prob

            stop = max(init_pos - Ld + delDr, min_pos)
            for pos in range(init_pos - 3, stop - 1, -1):
                D_pos = Ld - delDr - 1 - ((init_pos - 1) - pos)
                ok = _update_Pi_R(
                    D_pos, max_del_D, delDr, Pdel_D, aa_idx_seq,
                    pos, dseq, allow1, allow2, allow3, Pi_R, PD_nt,
                    PD_2nd_stack, base_prob
                )
                if not ok:
                    break
  
    # Frame 2
    for init_pos in range(-2, -L3 - 1, -3):
        _load4(Pij, v0, init_pos)
        for k in range(4):
            Pi_R[k, init_pos] += zeroD * v0[k]

        p_last = v0

        for delDr in range(num_delr):
            if min_del_D[delDr] == -1:
                continue

            dpos1 = Ld - delDr - 1
            last_idx = dseq[dpos1]
            base_prob = p_last[last_idx]
            if base_prob == 0.0:
                continue

            if dpos1 <= max_del_D[delDr]:
                Pi_R[0, init_pos - 1] += Pdel_D[dpos1, delDr] * base_prob

            stop = max(init_pos - Ld + delDr, min_pos)
            for pos in range(init_pos - 2, stop - 1, -1):
                D_pos = dpos1 - ((init_pos - 1) - pos)  
                ok = _update_Pi_R(
                    D_pos, max_del_D, delDr, Pdel_D, aa_idx_seq,
                    pos, dseq, allow1, allow2, allow3, Pi_R, PD_nt,
                    PD_2nd_stack, base_prob
                )
                if not ok:
                    break

    # Frame 3
    for init_pos in range(-3, -L3 - 1, -3):
        pA = Pij[0, init_pos]
        Pi_R[0, init_pos] += zeroD * pA
 
        base_prob = pA
        for delDr in range(num_delr):
            if min_del_D[delDr] == -1:
                continue

            stop = max(init_pos - Ld + delDr, min_pos)
            for pos in range(init_pos - 1, stop - 1, -1):
                D_pos = Ld - delDr - 1 - ((init_pos - 1) - pos)
                ok = _update_Pi_R(
                    D_pos, max_del_D, delDr, Pdel_D, aa_idx_seq,
                    pos, dseq, allow1, allow2, allow3, Pi_R, PD_nt,
                    PD_2nd_stack, base_prob
                )
                if not ok:
                    break


@nb.njit(fastmath=True)
def _update_Pi_R(
    D_pos: int,
    max_del_D: np.ndarray,
    delDr: int,
    Pdel_D: np.ndarray,
    aa_idx_seq: np.ndarray,
    pos: int,
    dseq: np.ndarray,
    allow1: np.ndarray,
    allow2: np.ndarray,
    allow3: np.ndarray,
    Pi_R: np.ndarray,
    PD_nt: np.ndarray,
    PD_2nd_stack: np.ndarray,
    base_prob: float
) -> bool:

    if D_pos > max_del_D[delDr]:
        current_PdelDldelDr = 0
    else:
        current_PdelDldelDr = Pdel_D[D_pos, delDr]

    aa = aa_idx_seq[pos // 3]
    r = pos % 3
    if r == 2:
        single = dseq[D_pos]
        if allow1[aa, single] != 0:
            w = current_PdelDldelDr * base_prob
            Pi_R[0, pos] += w * PD_nt[0, D_pos]
            Pi_R[1, pos] += w * PD_nt[1, D_pos]
            Pi_R[2, pos] += w * PD_nt[2, D_pos]
            Pi_R[3, pos] += w * PD_nt[3, D_pos]
            return True
    elif r == 1:
        pair = 4*dseq[D_pos] + dseq[D_pos + 1]
        if allow2[aa, pair] != 0:
            w = current_PdelDldelDr * base_prob
            Pi_R[0, pos] += w * PD_2nd_stack[aa, 0, D_pos]
            Pi_R[1, pos] += w * PD_2nd_stack[aa, 1, D_pos]
            Pi_R[2, pos] += w * PD_2nd_stack[aa, 2, D_pos]
            Pi_R[3, pos] += w * PD_2nd_stack[aa, 3, D_pos]
            return True
    else:
        triplet = 16*dseq[D_pos] + 4*dseq[D_pos + 1] + dseq[D_pos + 2]
        if allow3[aa, triplet] != 0:
            Pi_R[0, pos] += current_PdelDldelDr * base_prob
            return True

    return False


@nb.njit(inline='always')
def _load4(src: np.ndarray, dst: np.ndarray, init_pos: int) -> None:
    """ Load a integer into a 4-length array"""
    dst[0] = src[0, init_pos]
    dst[1] = src[1, init_pos]
    dst[2] = src[2, init_pos]
    dst[3] = src[3, init_pos]


@nb.njit(inline='always')
def _dot4(M: np.ndarray, x: np.ndarray, out: np.ndarray) -> None:
    """ Dot product of two 4-length arrays to avoid temp array
    allocation. Saves to existing third array `out`.
    """

    out[0] = M[0,0]*x[0] + M[0,1]*x[1] + M[0,2]*x[2] + M[0,3]*x[3]
    out[1] = M[1,0]*x[0] + M[1,1]*x[1] + M[1,2]*x[2] + M[1,3]*x[3]
    out[2] = M[2,0]*x[0] + M[2,1]*x[1] + M[2,2]*x[2] + M[2,3]*x[3]
    out[3] = M[3,0]*x[0] + M[3,1]*x[1] + M[3,2]*x[2] + M[3,3]*x[3]


@nb.njit(inline='always')
def _dot4_inplace(M: np.ndarray, x: np.ndarray) -> None:
    """ Dot product of two 4-length arrays to avoid temp array
    allocation. Rewrites to `x`.
    """

    a = M[0,0]*x[0] + M[0,1]*x[1] + M[0,2]*x[2] + M[0,3]*x[3]
    b = M[1,0]*x[0] + M[1,1]*x[1] + M[1,2]*x[2] + M[1,3]*x[3]
    c = M[2,0]*x[0] + M[2,1]*x[1] + M[2,2]*x[2] + M[2,3]*x[3]
    d = M[3,0]*x[0] + M[3,1]*x[1] + M[3,2]*x[2] + M[3,3]*x[3]
    x[0] = a; x[1] = b; x[2] = c; x[3] = d


@nb.njit(inline='always')
def _axpy4(scale: float, src: np.ndarray, dst: np.ndarray, col: int) -> None:
    """ For 4-length array, dst[:, col] += scale * src """

    dst[0, col] += scale * src[0]
    dst[1, col] += scale * src[1]
    dst[2, col] += scale * src[2]
    dst[3, col] += scale * src[3]


@nb.njit(inline='always')
def _dot4_axpy4(scale: float, M: np.ndarray, x: np.ndarray, dst: np.ndarray, col: int) -> None:
    """ dot4 -> axpy4 without temp allocation """

    a = M[0,0]*x[0] + M[0,1]*x[1] + M[0,2]*x[2] + M[0,3]*x[3]
    b = M[1,0]*x[0] + M[1,1]*x[1] + M[1,2]*x[2] + M[1,3]*x[3]
    c = M[2,0]*x[0] + M[2,1]*x[1] + M[2,2]*x[2] + M[2,3]*x[3]
    d = M[3,0]*x[0] + M[3,1]*x[1] + M[3,2]*x[2] + M[3,3]*x[3]

    dst[0, col] += scale * a
    dst[1, col] += scale * b
    dst[2, col] += scale * c
    dst[3, col] += scale * d


@nb.njit(inline='always')
def _sum4(x: np.ndarray) -> float:
    """ Sum a 4-length array """
    return x[0] + x[1] + x[2] + x[3]



"""Tests for sequence generation functionality."""
import pytest
import numpy as np
import olga.sequence_generation as sequence_generation
from olga.utils import nt2aa


class TestSequenceGenerationVDJ:
    """Tests for VDJ sequence generation."""
    
    def test_generate_sequence(self, human_t_beta_model):
        """Test generating a single sequence."""
        generative_model = human_t_beta_model['generative_model']
        genomic_data = human_t_beta_model['genomic_data']
        
        seq_gen = sequence_generation.SequenceGenerationVDJ(
            generative_model, genomic_data
        )
        
        ntseq, aaseq, V_in, J_in = seq_gen.gen_rnd_prod_CDR3()
        
        assert isinstance(ntseq, str)
        assert isinstance(aaseq, str)
        assert isinstance(V_in, (int, np.integer))
        assert isinstance(J_in, (int, np.integer))
        assert len(ntseq) > 0
        assert len(aaseq) > 0
        assert len(ntseq) % 3 == 0  # Must be in frame
        assert aaseq[0] == 'C'  # Should start with C
        assert aaseq[-1] in 'FVW'  # Should end with F, V, or W
        assert '*' not in aaseq  # Should be productive (no stop codons)
    
    def test_generate_multiple_sequences(self, human_t_beta_model):
        """Test generating multiple sequences."""
        generative_model = human_t_beta_model['generative_model']
        genomic_data = human_t_beta_model['genomic_data']
        
        seq_gen = sequence_generation.SequenceGenerationVDJ(
            generative_model, genomic_data
        )
        
        sequences = []
        for _ in range(10):
            ntseq, aaseq, V_in, J_in = seq_gen.gen_rnd_prod_CDR3()
            sequences.append((ntseq, aaseq, V_in, J_in))
            assert len(ntseq) > 0
            assert len(aaseq) > 0
        
        # Check that we got some variety
        unique_seqs = set(aaseq for _, aaseq, _, _ in sequences)
        assert len(unique_seqs) > 1  # Should have some variety
    
    def test_conserved_j_residues(self, human_t_beta_model):
        """Test with custom conserved J residues."""
        generative_model = human_t_beta_model['generative_model']
        genomic_data = human_t_beta_model['genomic_data']
        
        seq_gen = sequence_generation.SequenceGenerationVDJ(
            generative_model, genomic_data
        )
        
        ntseq, aaseq, V_in, J_in = seq_gen.gen_rnd_prod_CDR3(conserved_J_residues='F')
        
        assert aaseq[-1] == 'F'  # Should end with F only
    
    def test_sequence_consistency(self, human_t_beta_model):
        """Test that generated nucleotide and amino acid sequences are consistent."""
        generative_model = human_t_beta_model['generative_model']
        genomic_data = human_t_beta_model['genomic_data']
        
        seq_gen = sequence_generation.SequenceGenerationVDJ(
            generative_model, genomic_data
        )
        
        ntseq, aaseq, V_in, J_in = seq_gen.gen_rnd_prod_CDR3()
        
        # Check that translation is consistent (allowing for ambiguous codons)
        translated = nt2aa(ntseq)
        # The translated sequence should match or be compatible with aaseq
        assert len(translated) == len(aaseq) or len(translated) == len(ntseq) // 3


class TestSequenceGenerationVJ:
    """Tests for VJ sequence generation."""
    
    def test_generate_sequence(self, human_t_alpha_model):
        """Test generating a single VJ sequence."""
        generative_model = human_t_alpha_model['generative_model']
        genomic_data = human_t_alpha_model['genomic_data']
        
        seq_gen = sequence_generation.SequenceGenerationVJ(
            generative_model, genomic_data
        )
        
        ntseq, aaseq, V_in, J_in = seq_gen.gen_rnd_prod_CDR3()
        
        assert isinstance(ntseq, str)
        assert isinstance(aaseq, str)
        assert isinstance(V_in, (int, np.integer))
        assert isinstance(J_in, (int, np.integer))
        assert len(ntseq) > 0
        assert len(aaseq) > 0
        assert len(ntseq) % 3 == 0  # Must be in frame
        assert aaseq[0] == 'C'  # Should start with C
        assert aaseq[-1] in 'FVW'  # Should end with F, V, or W
        assert '*' not in aaseq  # Should be productive
    
    def test_generate_multiple_sequences(self, human_t_alpha_model):
        """Test generating multiple VJ sequences."""
        generative_model = human_t_alpha_model['generative_model']
        genomic_data = human_t_alpha_model['genomic_data']
        
        seq_gen = sequence_generation.SequenceGenerationVJ(
            generative_model, genomic_data
        )
        
        sequences = []
        for _ in range(10):
            ntseq, aaseq, V_in, J_in = seq_gen.gen_rnd_prod_CDR3()
            sequences.append((ntseq, aaseq, V_in, J_in))
        
        # Check variety
        unique_seqs = set(aaseq for _, aaseq, _, _ in sequences)
        assert len(unique_seqs) > 1


class TestSequenceGenerationAttributes:
    """Tests for sequence generation attributes."""
    
    def test_vdj_attributes(self, human_t_beta_model):
        """Test VDJ sequence generator has required attributes."""
        generative_model = human_t_beta_model['generative_model']
        genomic_data = human_t_beta_model['genomic_data']
        
        seq_gen = sequence_generation.SequenceGenerationVDJ(
            generative_model, genomic_data
        )
        
        assert hasattr(seq_gen, 'CPV'), f"CPV attribute not found in {seq_gen}"
        assert hasattr(seq_gen, 'CPDJ'), f"CPDJ attribute not found in {seq_gen}"
        assert hasattr(seq_gen, 'CinsVD'), f"CinsVD attribute not found in {seq_gen}"  
        assert hasattr(seq_gen, 'CinsDJ'), f"CinsDJ attribute not found in {seq_gen}"
        assert hasattr(seq_gen, 'cutV_genomic_CDR3_segs'), f"cutV_genomic_CDR3_segs attribute not found in {seq_gen}"
        assert hasattr(seq_gen, 'cutD_genomic_CDR3_segs'), f"cutD_genomic_CDR3_segs attribute not found in {seq_gen}"
        assert hasattr(seq_gen, 'cutJ_genomic_CDR3_segs'), f"cutJ_genomic_CDR3_segs attribute not found in {seq_gen}"
    
    def test_vj_attributes(self, human_t_alpha_model):
        """Test VJ sequence generator has required attributes."""
        generative_model = human_t_alpha_model['generative_model']
        genomic_data = human_t_alpha_model['genomic_data']
        
        seq_gen = sequence_generation.SequenceGenerationVJ(
            generative_model, genomic_data
        )
        
        assert hasattr(seq_gen, 'CPVJ'), f"CPVJ attribute not found in {seq_gen}"
        assert hasattr(seq_gen, 'CPinsVJ'), f"CPinsVJ attribute not found in {seq_gen}"
        assert hasattr(seq_gen, 'C_Rvj'), f"C_Rvj attribute not found in {seq_gen}"
        assert hasattr(seq_gen, 'C_first_nt_bias_insVJ'), f"C_first_nt_bias_insVJ attribute not found in {seq_gen}"
        assert hasattr(seq_gen, 'cutV_genomic_CDR3_segs'), f"cutV_genomic_CDR3_segs attribute not found in {seq_gen}"
        assert hasattr(seq_gen, 'cutJ_genomic_CDR3_segs'), f"cutJ_genomic_CDR3_segs attribute not found in {seq_gen}"




#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Command line script to generate sequences.

    Copyright (C) 2018 Zachary Sethna

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

This program will generate a file of Monte Carlo sampling from a specified
generative V(D)J model. The sequences generated will have NO ERRORS.

There are four default generative models that ship with OLGA and can be
specified with a flag:
--humanTRA (Human T cell alpha chain VJ model)
--humanTRB (Human T cell beta chain VDJ model)
--mouseTRB (Mouse T cell beta chain VDJ model)
--mouseTRA (Mouse T cell alpha chain VJ model)
--humanIGH (Human B cell heavy chain VDJ model)
--humanIGK (Human B cell light kappa chain VJ model)
--humanIGL (Human B cell light lambda chain VJ model)
--mouseIGH (Mouse B cell heavy chain VDJ model)
--mouseIGK (Mouse B cell light kappa chain VJ model)
--mouseIGL (Mouse B cell light lambda chain VJ model)

To specify a custom model folder use:
--set_custom_model_VJ (generative model of VJ recombination, e.g. T alpha chain)
--set_custom_model_VDJ (generative model of VDJ recombination, e.g. T beta chain)

Note, if specifying a custom model folder for either a VJ recombination model
(e.g. an alpha or light chain model) or a VDJ recombination model
(e.g. a beta or heavy chain model), the folder must contain the following files
with the exact naming convention:

model_params.txt (IGoR inference param file)
model_marginals.txt (IGoR inference marginal file)
V_gene_CDR3_anchors.csv (V residue anchor and functionality file)
J_gene_CDR3_anchors.csv (J residue anchor and functionality file)


It is required to specify the number of sequences to be generated. This is done
with -n (see Options).

If a file is specified to write to (using -o, see Options), the generated
sequences are written to the file, otherwise they are printed to stdout.

The default is to record both the nucleotide CDR3 sequence and the amino acid
CDR3 sequence. This can be specified (see Options).

The V/J genes used to generate each sequence can be recorded or not. Default is
to record them, but this can be toggled off with --record_genes_off (see Options)

-------------------------------------------------------------------------------
Example calls:

#Print 20 generated sequences to stdout
$ olga-generate_sequences --humanTRB -n 20

#Write the 200 generated sequences to example_seqs.tsv
$ olga-generate_sequences --humanTRB -o example_seqs.tsv -n 200

#Write 20,000 generated sequences to example_seqs.tsv
$ olga-generate_sequences --humanTRB -o example_seqs.tsv -n 2e4

#Write only the amino acid sequences
$ olga-generate_sequences --humanTRB -o example_seqs.tsv -n 200 --seq_type amino_acid

--------------------------------------------------------------------------------
Options:
  -h, --help            show this help message and exit
  --humanTRA, --human_T_alpha
                        use default human TRA model (T cell alpha chain)
  --humanTRB, --human_T_beta
                        use default human TRB model (T cell beta chain)
  --mouseTRB, --mouse_T_beta
                        use default mouse TRB model (T cell beta chain)
  --mouseTRA, --mouse_T_alpha
                        use default mouse TRA model (T cell alpha chain)
  --humanIGH, --human_B_heavy
                        use default human IGH model (B cell heavy chain)
  --humanIGK
                        use default human IGK model (B cell light kappa chain)
  --humanIGL, --human_B_lambda
                        use default human IGL model (B cell light lambda chain) 
  --mouseIGH, --mouse_B_heavy
                        use default mouse IGH model (B cell heavy chain)
  --mouseIGK, --mouse_B_kappa
                        use default mouse IGK model (B cell light kappa chain)
  --mouseIGL, --mouse_B_lambda
                        use default mouse IGL model (B cell light lambda chain)
  --VDJ_model_folder=PATH/TO/FOLDER/
                        specify PATH/TO/FOLDER/ for a custom VDJ generative
                        model
  --VJ_model_folder=PATH/TO/FOLDER/
                        specify PATH/TO/FOLDER/ for a custom VJ generative
                        model
  -o PATH/TO/FILE, --outfile=PATH/TO/FILE
                        write CDR3 sequences to PATH/TO/FILE
  -n N, --num_seqs=N    specify the number of sequences to generate.
  --seed=SEED           set seed for pseudorandom number generator. Default is
                        to not set a seed.
  --seqs_per_time_update=SEQS_PER_TIME_UPDATE
                        specify the number of sequences between time updates.
                        Default is 1e5
  --conserved_J_residues=CONSERVED_J_RESIDUES
                        specify conserved J residues. Default is 'FVW'.
  --time_updates_off    turn time updates off.
  --seq_type=SEQ_TYPE   declare sequence type for output sequences. Choices:
                        'all' [default], 'ntseq', 'nucleotide', 'aaseq',
                        'amino_acid'
  --record_genes_off    turn off recording V and J gene info.
  -d DELIMITER, --delimiter=DELIMITER
                        declare delimiter choice. Default is tab for .tsv
                        output files, comma for .csv files, and tab for all
                        others. Choices: 'tab', 'space', ',', ';', ':'
  --raw_delimiter=DELIMITER
                        declare delimiter choice as a raw string.


Note about conserved_J_residues:

This specifies a string which must be composed ONLY of amino acids
(i.e. only ACDEFGHIKLMNPQRSTVWY*). The amino acids in that string will
determine functionality of a sequence. Please note that the J genes are
ALREADY ANCHORED at a given residue, thus this string should almost
certainly only include phenylalanine (F) and/or tryptophan (W). If amino
acids are used to define this 'conserved residue' in the string here, but
the J genes are still anchored at a 'F' or 'W' (as the default genomic
files are), this could result in no productive sequences being generated.
Unless the anchor positions are changed, LEAVE THE DEFAULT. The default
string is 'FVW'.

--------------------------------------------------------------------------------

@author: zacharysethna

"""

#Function assumes that it is in the same directory that the folder app/ is
#in (which should contain all the modules imported).
from __future__ import print_function, division
import os
import sys
import subprocess
reqs = subprocess.check_output([sys.executable, '-m', 'pip', 'freeze'])
installed_packages = [r.decode().split('==')[0] for r in reqs.split()]
if 'olga' not in installed_packages: sys.path.insert(0, os.path.split(os.path.dirname(os.path.abspath(__file__)))[0])
import olga.load_model as load_model
import olga.sequence_generation as sequence_generation
#import load_model as load_model
#import sequence_generation as sequence_generation
from optparse import OptionParser
import time
import numpy as np

#Set input = raw_input for python 2
try:
    import __builtin__
    input = getattr(__builtin__, 'raw_input')
except (ImportError, AttributeError):
    pass

def main():
    """ Generate sequences."""

    parser = OptionParser(conflict_handler="resolve")

    parser.add_option('--humanTRA', '--human_T_alpha', action='store_true', dest='humanTRA', default=False, help='use default human TRA model (T cell alpha chain)')
    parser.add_option('--humanTRB', '--human_T_beta', action='store_true', dest='humanTRB', default=False, help='use default human TRB model (T cell beta chain)')
    parser.add_option('--mouseTRB', '--mouse_T_beta', action='store_true', dest='mouseTRB', default=False, help='use default mouse TRB model (T cell beta chain)')
    parser.add_option('--mouseTRA', '--mouse_T_alpha', action='store_true', dest='mouseTRA', default=False, help='use default mouse TRA model (T cell alpha chain)')
    parser.add_option('--humanIGH', '--human_B_heavy', action='store_true', dest='humanIGH', default=False, help='use default human IGH model (B cell heavy chain)')
    parser.add_option('--humanIGK', '--human_B_kappa', action='store_true', dest='humanIGK', default=False, help='use default human IGK model (B cell light kappa chain)')
    parser.add_option('--humanIGL', '--human_B_lambda', action='store_true', dest='humanIGL', default=False, help='use default human IGL model (B cell light lambda chain)')
    parser.add_option('--mouseIGH', '--mouse_B_heavy', action='store_true', dest='mouseIGH', default=False, help='use default mouse IGH model (B cell heavy chain)')
    parser.add_option('--mouseIGK', '--mouse_B_kappa', action='store_true', dest='mouseIGK', default=False, help='use default mouse IGK model (B cell light kappa chain)')
    parser.add_option('--mouseIGL', '--mouse_B_lambda', action='store_true', dest='mouseIGL', default=False, help='use default mouse IGL model (B cell light lambda chain)')
    parser.add_option('--VDJ_model_folder','--set_custom_model_VDJ', dest='vdj_model_folder', metavar='PATH/TO/FOLDER/', help='specify PATH/TO/FOLDER/ for a custom VDJ generative model')
    parser.add_option('--VJ_model_folder','--set_custom_model_VJ', dest='vj_model_folder', metavar='PATH/TO/FOLDER/', help='specify PATH/TO/FOLDER/ for a custom VJ generative model')
    parser.add_option('-o', '--outfile', dest = 'outfile_name', metavar='PATH/TO/FILE', help='write CDR3 sequences to PATH/TO/FILE')
    parser.add_option('-n', '--num_seqs', type='float', metavar='N', default = 0, dest='num_seqs_to_generate', help='specify the number of sequences to generate.')
    parser.add_option('--seed', type='int', dest='seed', help='set seed for pseudorandom number generator. Default is to not set a seed.')
    parser.add_option('--seqs_per_time_update', type='float', default = 100000, dest='seqs_per_time_update', help='specify the number of sequences between time updates. Default is 1e5')
    parser.add_option('--conserved_J_residues', type='string', default = 'FVW', dest='conserved_J_residues', help="specify conserved J residues. Default is 'FVW'.")
    parser.add_option('--time_updates_off', action='store_false', dest='time_updates', default=True, help='turn time updates off.')
    parser.add_option('--seq_type', type='choice', default = 'all', dest='seq_type',  choices=['all', 'ntseq', 'nucleotide', 'aaseq', 'amino_acid'], help="declare sequence type for output sequences. Choices: 'all' [default], 'ntseq', 'nucleotide', 'aaseq', 'amino_acid'")
    parser.add_option('--record_genes_off', action='store_false', dest="record_genes", default=True, help='turn off recording V and J gene info.')
    parser.add_option('-d', '--delimiter', type='choice', dest='delimiter',  choices=['tab', 'space', ',', ';', ':'], help="declare delimiter choice. Default is tab for .tsv output files, comma for .csv files, and tab for all others. Choices: 'tab', 'space', ',', ';', ':'")
    parser.add_option('--raw_delimiter', type='str', dest='delimiter', help="declare delimiter choice as a raw string.")


    (options, args) = parser.parse_args()

    main_folder = os.path.dirname(__file__)

    default_models = {}
    default_models['humanTRA'] = [os.path.join(main_folder, 'default_models', 'human_T_alpha'),  'VJ']
    default_models['humanTRB'] = [os.path.join(main_folder, 'default_models', 'human_T_beta'), 'VDJ']
    default_models['mouseTRB'] = [os.path.join(main_folder, 'default_models', 'mouse_T_beta'), 'VDJ']
    default_models['mouseTRA'] = [os.path.join(main_folder, 'default_models', 'mouse_T_alpha'), 'VJ']
    default_models['humanIGH'] = [os.path.join(main_folder, 'default_models', 'human_B_heavy'), 'VDJ']
    default_models['humanIGK'] = [os.path.join(main_folder, 'default_models', 'human_B_kappa'), 'VJ']
    default_models['humanIGL'] = [os.path.join(main_folder, 'default_models', 'human_B_lambda'),  'VJ']
    default_models['mouseIGH'] = [os.path.join(main_folder, 'default_models', 'mouse_B_heavy'), 'VDJ']
    default_models['mouseIGK'] = [os.path.join(main_folder, 'default_models', 'mouse_B_kappa'), 'VJ']
    default_models['mouseIGL'] = [os.path.join(main_folder, 'default_models', 'mouse_B_lambda'), 'VJ']
    num_models_specified = sum([1 for x in list(default_models.keys()) + ['vj_model_folder', 'vdj_model_folder'] if getattr(options, x)])

    if num_models_specified == 1: #exactly one model specified
        try:
            d_model = [x for x in default_models.keys() if getattr(options, x)][0]
            model_folder = default_models[d_model][0]
            recomb_type = default_models[d_model][1]
        except IndexError:
            if options.vdj_model_folder: #custom VDJ model specified
                model_folder = options.vdj_model_folder
                recomb_type = 'VDJ'
            elif options.vj_model_folder: #custom VJ model specified
                model_folder = options.vj_model_folder
                recomb_type = 'VJ'
    elif num_models_specified == 0:
        print('Need to indicate generative model.')
        print('Exiting...')
        return -1
    elif num_models_specified > 1:
        print('Only specify one model')
        print('Exiting...')
        return -1

    #Check that all model and genomic files exist in the indicated model folder
    if not os.path.isdir(model_folder):
        print('Check pathing... cannot find the model folder: ' + model_folder)
        print('Exiting...')
        return -1

    params_file_name = os.path.join(model_folder,'model_params.txt')
    marginals_file_name = os.path.join(model_folder,'model_marginals.txt')
    V_anchor_pos_file = os.path.join(model_folder,'V_gene_CDR3_anchors.csv')
    J_anchor_pos_file = os.path.join(model_folder,'J_gene_CDR3_anchors.csv')

    for x in [params_file_name, marginals_file_name, V_anchor_pos_file, J_anchor_pos_file]:
        if not os.path.isfile(x):
            print('Cannot find: ' + x)
            print('Please check the files (and naming conventions) in the model folder ' + model_folder)
            print('Exiting...')
            return -1


    if options.outfile_name is not None:
        outfile_name = options.outfile_name
        if os.path.isfile(outfile_name):
            if not input(outfile_name + ' already exists. Overwrite (y/n)? ').strip().lower() in ['y', 'yes']:
                print('Exiting...')
                return -1

    #Parse arguments

    num_seqs_to_generate = int(options.num_seqs_to_generate)

    if num_seqs_to_generate <= 0:
        print('Need to specify num_seqs (number of sequences to generate).')
        print('Exiting...')
        return -1

    #Parse default delimiter
    delimiter = options.delimiter
    if delimiter is None:
        delimiter = '\t'
        if options.outfile_name is not None:
            if outfile_name.endswith('.tsv'):
                delimiter = '\t'
            elif outfile_name.endswith('.csv'):
                delimiter = ','
    else:
        try:
            delimiter = {'tab': '\t', 'space': ' ', ',': ',', ';': ';', ':': ':'}[delimiter]
        except KeyError:
            pass #Other raw string.

    #Optional flags
    seq_type = {'all': 'all', 'ntseq': 'ntseq', 'nucleotide': 'ntseq', 'aaseq': 'aaseq', 'amino_acid': 'aaseq'}[options.seq_type]
    record_genes = options.record_genes
    seqs_per_time_update = int(options.seqs_per_time_update)
    time_updates = options.time_updates
    conserved_J_residues = options.conserved_J_residues

    if options.seed is not None:
        np.random.seed(options.seed)

    #VDJ recomb case --- used for TCRB and IGH
    if recomb_type == 'VDJ':
        genomic_data = load_model.GenomicDataVDJ()
        genomic_data.load_igor_genomic_data(params_file_name, V_anchor_pos_file, J_anchor_pos_file)
        generative_model = load_model.GenerativeModelVDJ()
        generative_model.load_and_process_igor_model(marginals_file_name)
        seq_gen = sequence_generation.SequenceGenerationVDJ(generative_model, genomic_data)
    #VJ recomb case --- used for TCRA and light chain
    elif recomb_type == 'VJ':
        genomic_data = load_model.GenomicDataVJ()
        genomic_data.load_igor_genomic_data(params_file_name, V_anchor_pos_file, J_anchor_pos_file)
        generative_model = load_model.GenerativeModelVJ()
        generative_model.load_and_process_igor_model(marginals_file_name)
        seq_gen = sequence_generation.SequenceGenerationVJ(generative_model, genomic_data)


    V_gene_names = [V[0].split('*')[0] for V in genomic_data.genV]
    J_gene_names = [J[0].split('*')[0] for J in genomic_data.genJ]

    if options.outfile_name is not None:
        outfile = open(outfile_name, 'w')

        print('Starting sequence generation... ')
        start_time = time.time()
        for i in range(num_seqs_to_generate):
            ntseq, aaseq, V_in, J_in = seq_gen.gen_rnd_prod_CDR3(conserved_J_residues)
            if seq_type == 'all': #default, include both ntseq and aaseq
                current_line_out = ntseq + delimiter + aaseq
            elif seq_type == 'ntseq': #only record ntseq
                current_line_out = ntseq
            elif seq_type == 'aaseq': #only record aaseq
                current_line_out = aaseq

            if record_genes:
                current_line_out += delimiter + V_gene_names[V_in] + delimiter + J_gene_names[J_in]
            outfile.write(current_line_out + '\n')

            if (i+1)%seqs_per_time_update == 0 and time_updates:
                c_time = time.time() - start_time
                eta = ((num_seqs_to_generate - (i+1))/(i+1))*c_time
                if c_time > 86400: #more than a day
                    c_time_str = '%d days, %d hours, %d minutes, and %.2f seconds.'%(int(c_time)//86400, (int(c_time)//3600)%24, (int(c_time)//60)%60, c_time%60)
                elif c_time > 3600: #more than an hr
                    c_time_str = '%d hours, %d minutes, and %.2f seconds.'%((int(c_time)//3600)%24, (int(c_time)//60)%60, c_time%60)
                elif c_time > 60: #more than a min
                    c_time_str = '%d minutes and %.2f seconds.'%((int(c_time)//60)%60, c_time%60)
                else:
                    c_time_str = '%.2f seconds.'%(c_time)

                if eta > 86400: #more than a day
                    eta_str = '%d days, %d hours, %d minutes, and %.2f seconds.'%(int(eta)//86400, (int(eta)//3600)%24, (int(eta)//60)%60, eta%60)
                elif eta > 3600: #more than an hr
                    eta_str = '%d hours, %d minutes, and %.2f seconds.'%((int(eta)//3600)%24, (int(eta)//60)%60, eta%60)
                elif eta > 60: #more than a min
                    eta_str = '%d minutes and %.2f seconds.'%((int(eta)//60)%60, eta%60)
                else:
                    eta_str = '%.2f seconds.'%(eta)

                print('%d sequences generated in %s Estimated time remaining: %s'%(i+1, c_time_str, eta_str))

        c_time = time.time() - start_time
        if c_time > 86400: #more than a day
            c_time_str = '%d days, %d hours, %d minutes, and %.2f seconds.'%(int(c_time)//86400, (int(c_time)//3600)%24, (int(c_time)//60)%60, c_time%60)
        elif c_time > 3600: #more than an hr
            c_time_str = '%d hours, %d minutes, and %.2f seconds.'%((int(c_time)//3600)%24, (int(c_time)//60)%60, c_time%60)
        elif c_time > 60: #more than a min
            c_time_str = '%d minutes and %.2f seconds.'%((int(c_time)//60)%60, c_time%60)
        else:
            c_time_str = '%.2f seconds.'%(c_time)
        print('Completed generating all %d sequences in %s'%(num_seqs_to_generate, c_time_str))
        outfile.close()

    else: #print to stdout
        for i in range(num_seqs_to_generate):
            ntseq, aaseq, V_in, J_in = seq_gen.gen_rnd_prod_CDR3(conserved_J_residues)
            if seq_type == 'all': #default, include both ntseq and aaseq
                current_line_out = ntseq + delimiter + aaseq
            elif seq_type == 'ntseq': #only record ntseq
                current_line_out = ntseq
            elif seq_type == 'aaseq': #only record aaseq
                current_line_out = aaseq

            if record_genes:
                current_line_out += delimiter + V_gene_names[V_in] + delimiter + J_gene_names[J_in]
            print(current_line_out)

if __name__ == '__main__': main()



"""Tests for FastPgen (numba-accelerated) functionality."""
import pytest
import numpy as np
from olga.performance.fast_pgen import FastPgen


class TestFastPgenVDJ:
    """Tests for FastPgen with VDJ models."""
    
    def test_fast_pgen_initialization(self, human_t_beta_model):
        """Test FastPgen initialization with VDJ model."""
        pgen_model = human_t_beta_model['pgen_model']
        
        fast_model = FastPgen(pgen_model)
        
        assert fast_model._impl == pgen_model
        assert hasattr(fast_model, 'codons_dict')
    
    def test_fast_pgen_compute_aa(self, human_t_beta_model):
        """Test FastPgen amino acid computation."""
        pgen_model = human_t_beta_model['pgen_model']
        fast_model = FastPgen(pgen_model)
        
        cdr3_seq = 'CASSTGQANYGYTF'
        
        pgen_slow = pgen_model.compute_aa_CDR3_pgen(cdr3_seq)
        pgen_fast = fast_model.compute_aa_CDR3_pgen(cdr3_seq)
        
        # Results should be very close (within numerical precision)
        assert np.allclose(pgen_slow, pgen_fast, rtol=1e-6)
        assert pgen_fast > 0
    
    def test_fast_pgen_compute_nt(self, human_t_beta_model):
        """Test FastPgen nucleotide computation."""
        pgen_model = human_t_beta_model['pgen_model']
        fast_model = FastPgen(pgen_model)
        
        ntseq = 'TGTGCCAGCAGTGACGCACAGGGGCGTAATCGTGGGACTGAAGCTTTCTTT'
        
        pgen_slow = pgen_model.compute_nt_CDR3_pgen(ntseq, print_warnings=False)
        pgen_fast = fast_model.compute_nt_CDR3_pgen(ntseq, print_warnings=False)
        
        # Results should be very close
        assert np.allclose(pgen_slow, pgen_fast, rtol=1e-6)
    
    def test_fast_pgen_with_masks(self, human_t_beta_model):
        """Test FastPgen with V and J masks."""
        pgen_model = human_t_beta_model['pgen_model']
        fast_model = FastPgen(pgen_model)
        
        cdr3_seq = 'CASSTGQANYGYTF'
        
        # Get some V and J genes
        v_genes = list(pgen_model.V_allele_names[:2]) if len(pgen_model.V_allele_names) >= 2 else pgen_model.V_allele_names
        j_genes = list(pgen_model.J_allele_names[:2]) if len(pgen_model.J_allele_names) >= 2 else pgen_model.J_allele_names
        
        pgen_slow = pgen_model.compute_aa_CDR3_pgen(cdr3_seq, v_genes, j_genes)
        pgen_fast = fast_model.compute_aa_CDR3_pgen(cdr3_seq, v_genes, j_genes)
        
        assert np.allclose(pgen_slow, pgen_fast, rtol=1e-6)


class TestFastPgenVJ:
    """Tests for FastPgen with VJ models."""
    
    def test_fast_pgen_initialization(self, human_t_alpha_model):
        """Test FastPgen initialization with VJ model."""
        pgen_model = human_t_alpha_model['pgen_model']
        
        fast_model = FastPgen(pgen_model)
        
        assert fast_model._impl == pgen_model
    
    def test_fast_pgen_compute_aa(self, human_t_alpha_model):
        """Test FastPgen amino acid computation for VJ model."""
        pgen_model = human_t_alpha_model['pgen_model']
        fast_model = FastPgen(pgen_model)
        
        cdr3_seq = 'CAVKIQGAQKLVF'
        
        pgen_slow = pgen_model.compute_aa_CDR3_pgen(cdr3_seq)
        pgen_fast = fast_model.compute_aa_CDR3_pgen(cdr3_seq)
        
        assert np.allclose(pgen_slow, pgen_fast, rtol=1e-6)
        assert pgen_fast > 0


class TestFastPgenPickling:
    """Tests for FastPgen pickling (for multiprocessing)."""
    
    def test_fast_pgen_pickling(self, human_t_beta_model):
        """Test that FastPgen can be pickled and unpickled."""
        import pickle
        
        pgen_model = human_t_beta_model['pgen_model']
        fast_model = FastPgen(pgen_model)
        
        # Test that it can be pickled
        pickled = pickle.dumps(fast_model)
        unpickled = pickle.loads(pickled)
        
        # Test that unpickled model still works
        cdr3_seq = 'CASSTGQANYGYTF'
        pgen_original = fast_model.compute_aa_CDR3_pgen(cdr3_seq)
        pgen_unpickled = unpickled.compute_aa_CDR3_pgen(cdr3_seq)
        
        assert np.allclose(pgen_original, pgen_unpickled, rtol=1e-6)


class TestFastPgenConsistency:
    """Tests for FastPgen consistency with standard implementation."""
    
    def test_multiple_sequences_consistency(self, human_t_beta_model):
        """Test that FastPgen gives consistent results for multiple sequences."""
        pgen_model = human_t_beta_model['pgen_model']
        fast_model = FastPgen(pgen_model)
        
        test_sequences = [
            'CASSTGQANYGYTF',
            'CAWSVAPDRGGYTF',
            'CASSDAQGRNRGTEAFF'
        ]
        
        for seq in test_sequences:
            pgen_slow = pgen_model.compute_aa_CDR3_pgen(seq, print_warnings=False)
            pgen_fast = fast_model.compute_aa_CDR3_pgen(seq, print_warnings=False)
            
            if pgen_slow > 0:  # Only check if slow version gives valid result
                assert np.allclose(pgen_slow, pgen_fast, rtol=1e-6)

